from eth_tester import (
    EthereumTester as _EthereumTester,
)
from eth_utils import (
    is_hex,
)

from .utils.validation import (
    validate_address,
)


class EthereumTester(_EthereumTester):

    def sign_message(self, address, message):
        validate_address(address)
        assert is_hex(message) is True, "message needs to be of type hex"
        return self.backend.sign_message(address, message)



from eth_utils import (
    add_0x_prefix,
    apply_to_return_value,
    keccak as eth_utils_keccak,
    remove_0x_prefix,
    to_bytes,
)
from eth_utils.toolz import (
    curry,
)
from hexbytes import (
    HexBytes,
)

from .abi import (
    map_abi_data,
)
from .encoding import (
    hex_encode_abi_type,
)


@curry
def abi_ens_resolver(type_str, val):
    return type_str, val


@apply_to_return_value(HexBytes)
def keccak(primitive=None, text=None, hexstr=None):
    if isinstance(primitive, (bytes, int, type(None))):
        input_bytes = to_bytes(primitive, hexstr=hexstr, text=text)
        return eth_utils_keccak(input_bytes)

    raise TypeError(
        "You called keccak with first arg %r and keywords %r. You must call it with one of "
        "these approaches: keccak(text='txt'), keccak(hexstr='0x747874'), "
        "keccak(b'\\x74\\x78\\x74'), or keccak(0x747874)." % (
            primitive,
            {'text': text, 'hexstr': hexstr}
        )
    )


def solidityKeccak(abi_types, values):
    """
    Executes keccak256 exactly as Solidity does.
    Takes list of abi_types as inputs -- `[uint24, int8[], bool]`
    and list of corresponding values  -- `[20, [-1, 5, 0], True]`
    """
    if len(abi_types) != len(values):
        raise ValueError(
            "Length mismatch between provided abi types and values.  Got "
            "{0} types and {1} values.".format(len(abi_types), len(values))
        )

    normalized_values = map_abi_data([abi_ens_resolver], abi_types, values)

    hex_string = add_0x_prefix(''.join(
        remove_0x_prefix(hex_encode_abi_type(abi_type, value))
        for abi_type, value
        in zip(abi_types, normalized_values)
    ))
    return keccak(hexstr=hex_string)



"""
code copied from web3._utils.decorators
"""
import functools
import threading


def reject_recursive_repeats(to_wrap):
    """
    Prevent simple cycles by returning None when called recursively with same instance
    """
    to_wrap.__already_called = {}

    @functools.wraps(to_wrap)
    def wrapped(*args):
        arg_instances = tuple(map(id, args))
        thread_id = threading.get_ident()
        thread_local_args = (thread_id,) + arg_instances
        if thread_local_args in to_wrap.__already_called:
            raise ValueError('Recursively called %s with %r' % (to_wrap, args))
        to_wrap.__already_called[thread_local_args] = True
        try:
            wrapped_val = to_wrap(*args)
        finally:
            del to_wrap.__already_called[thread_local_args]
        return wrapped_val
    return wrapped



from eth_tester.backends.pyevm.main import (
    PyEVMBackend as _PyEVMBackend,
)
from eth_utils import (
    ValidationError,
    to_bytes,
)

from .utils.solidity import (
    solidityKeccak,
)


class PyEVMBackend(_PyEVMBackend):

    def sign_message(self, address, message):
        for key in self.account_keys:
            if key.public_key.to_checksum_address() == address:
                private_key = key
                break
        else:
            raise ValidationError(
                f'Could not find address: {address} in keychain.'
                f'Use personal_newAccount to add a new account with privateKey'
            )
        message_hash = to_bytes(hexstr=message)
        message_hash = solidityKeccak(
            ['string', 'bytes32'],
            ['\x19Ethereum Signed Message:\n32', message_hash]
        )
        return str(
            private_key.sign_msg_hash(message_hash)
        )



"""
code copied from web3._utils.encoding
"""
# String encodings and numeric representations
from eth_utils import (
    add_0x_prefix,
    encode_hex,
    is_bytes,
    remove_0x_prefix,
    to_hex,
)

from .abi import (
    is_address_type,
    is_array_type,
    is_bool_type,
    is_bytes_type,
    is_int_type,
    is_string_type,
    is_uint_type,
    size_of_type,
    sub_type_of_array_type,
)
from .validation import (
    validate_abi_type,
    validate_abi_value,
)


def hex_encode_abi_type(abi_type, value, force_size=None):
    """
    Encodes value into a hex string in format of abi_type
    """
    validate_abi_type(abi_type)
    validate_abi_value(abi_type, value)

    data_size = force_size or size_of_type(abi_type)
    if is_array_type(abi_type):
        sub_type = sub_type_of_array_type(abi_type)
        return "".join([remove_0x_prefix(hex_encode_abi_type(sub_type, v, 256)) for v in value])
    elif is_bool_type(abi_type):
        return to_hex_with_size(value, data_size)
    elif is_uint_type(abi_type):
        return to_hex_with_size(value, data_size)
    elif is_int_type(abi_type):
        return to_hex_twos_compliment(value, data_size)
    elif is_address_type(abi_type):
        return pad_hex(value, data_size)
    elif is_bytes_type(abi_type):
        if is_bytes(value):
            return encode_hex(value)
        else:
            return value
    elif is_string_type(abi_type):
        return to_hex(text=value)
    else:
        raise ValueError(
            "Unsupported ABI type: {0}".format(abi_type)
        )


def to_hex_twos_compliment(value, bit_size):
    """
    Converts integer value to twos compliment hex representation with given bit_size
    """
    if value >= 0:
        return to_hex_with_size(value, bit_size)

    value = (1 << bit_size) + value
    hex_value = hex(value)
    hex_value = hex_value.rstrip("L")
    return hex_value


def to_hex_with_size(value, bit_size):
    """
    Converts a value to hex with given bit_size:
    """
    return pad_hex(to_hex(value), bit_size)


def pad_hex(value, bit_size):
    """
    Pads a hex string up to the given bit_size
    """
    value = remove_0x_prefix(value)
    return add_0x_prefix(value.zfill(int(bit_size / 4)))
